package aStarLibrary;

// AStar Navigation class, Aaron Steed 2006

// Some code and structure still intact from Tom Carden's version:
// <http://www.tom-carden.co.uk/p5/a_star_web/applet/index.html>
// Some links that helped me:
// <http://www.policyalmanac.org/games/aStarTutorial.htm>
// <http://www.geocities.com/jheyesjones/astar.html>

import java.util.Vector;

public class AStar {

  public Vector nodes;                          // Storage Vector for the Nodes
  public Vector open = new Vector();            // Possible Nodes for consideration
  public Vector closed = new Vector();          // Best of the Nodes
  public boolean wrap = false;                  // Setting for makeCuboidWeb() for grid wrap around
  public boolean corners = false;               // Setting for makeCuboidWeb() for connecting nodes at corners
  public boolean manhattan = true;              // Setting for using Manhattan distance measuring method (false uses Euclidean method)
  public boolean bfs = false;                   // Turns algorithm into a best first search
  public float [] offset;                       // Setting for makeCuboidWeb(), values in array are added to all Nodes

  // Constructors

  public AStar(){
    this.nodes = new Vector();
  }

  public AStar(Vector nodes) {
    this.nodes = nodes;
  }

  public AStar(int [] dim, float scale){
    makeCuboidNodes(dim, scale);
  }

  // AStar algorithm by Tom Carden,
  // BFS algorithm by Aaron Steed

  public Vector getPath(Node start, Node finish){
    for(int i = 0; i < nodes.size(); i++){
      Node n = (Node)nodes.get(i);
      n.reset();
    }
    open.clear();
    closed.clear();
    open.add(start);
    while(open.size() > 0){
      float lowest = Float.MAX_VALUE;
      int c = -1;
      for(int i = 0; i < open.size(); i++){
        Node temp = (Node)open.get(i);
        if(temp.f < lowest){
          lowest = temp.f;
          c = i;
        }
      }
      Node current = (Node)open.remove(c);
      closed.add(current);
      if(current == finish){
        break;
      }
      for(int i = 0; i < current.links.size(); i++){
        Node adjacent = (Node)current.links.get(i);
        if(adjacent.walkable && !closed.contains(adjacent)){
          if(!open.contains(adjacent)){
            open.add(adjacent);
            adjacent.parent = current;
            if(manhattan) adjacent.MsetF(finish, bfs); 
            else adjacent.setF(finish, bfs);
          }
          else if (!bfs){
            if(adjacent.g > current.g + current.dist(adjacent.p)){
              adjacent.parent = current;
              if(manhattan) adjacent.MsetF(finish, bfs); 
              else adjacent.setF(finish, bfs);
            }
          }
        }
      }
    }
    // Path generation
    Vector path = new Vector();
    Node pathNode = finish;
    while(pathNode != null){
      path.add(pathNode);
      pathNode = pathNode.parent;
    }
    // Hack to provide a compromise path when a route to the finish node is unavailable
    Node test = (Node)path.get(path.size()-1);
    if(test == finish){
      float leastDist = Float.MAX_VALUE;
      Node bestNode = null;
      for(int i = 0; i < closed.size(); i++){
        Node n = (Node)closed.get(i);
        float nDist = n.dist(finish.p);
        if(nDist < leastDist){
          leastDist = nDist;
          bestNode = n;
        }
      }
      if(bestNode.parent != null){
        pathNode = bestNode;
        path = new Vector();
        while(pathNode != null){
          path.add(pathNode);
          pathNode = pathNode.parent;
        }
      }
    }
    return path;
  }

  // Shortcut to adding a makeCuboidWeb construct to AStar

  public void makeCuboidNodes(int [] dim, float scale){
    nodes = new Vector();
    nodes = makeCuboidWeb(dim, scale);
  }

  public void addNodes(Vector nodes){
    this.nodes.addAll(nodes);
  }

  // Creates a construct of Nodes and connects them across adjacent dimensions.
  // Adapts for corners and wrap-around but at a cost of speed - only for init

  public Vector makeCuboidWeb(int [] dim, float scale){
    Vector world = new Vector();
    int totalLength = 1;
    for(int i = 0; i < dim.length; i++){
      if(dim[i] > 0){
        totalLength *= dim[i];
      }
    }
    for(int i = 0; i < totalLength; i++){
      int [] intP = getFolded(i, dim);
      float [] p = new float[intP.length];
      for(int j = 0; j < p.length; j++){
        p[j] = intP[j] * scale;
      }
      Node temp = new Node(p);
      if(offset != null){
        temp.p = temp.addP(offset);
      }
      world.add(temp);
    }
    int directions = (int)Math.pow(3, dim.length);
    for(int i = 0; i < totalLength; i++){
      int [] p = getFolded(i, dim);
      Node myNode = (Node)world.get(i);
      int [] b = new int[p.length];
      int [] w = new int[p.length];
      for(int j = 0; j < b.length; j++){
        b[j] = p[j] - 1;
        w[j] = 0;
      }
      for(int j = 0; j < directions; j++){
        boolean valid = true;
        for(int k = 0; k < dim.length; k++){
          if(b[k] > dim[k] - 1 || b[k] < 0){
            if(wrap){
              if(b[k] > dim[k] - 1){
                b[k] -= dim[k];
                w[k]--;
              }
              if(b[k] < 0){
                b[k] += dim[k];
                w[k]++;
              }
            } 
            else {
              valid = false;
            }
          }
          if(!corners){
            int combinations = 0;
            for(int l = 0; l < dim.length; l++){
              if(b[l] != p[l]){
                combinations++;
              }
            }
            if(combinations > 1){
              valid = false;
            }
          }
        }
        if(valid){
          Node connectee = (Node)world.get(getUnfolded(b, dim));
          if(myNode != connectee){
            myNode.connect(connectee);
          }
        }
        if(wrap){
          for(int k = 0; k < dim.length; k++){
            switch(w[k]){
            case 1:
              b[k] -= dim[k];
              w[k] = 0;
              break;
            case -1:
              b[k] += dim[k];
              w[k] = 0;
              break;
            }
          }
        }
        b[0]++;
        for(int k = 0; k < b.length - 1; k++){
          if(b[k] > p[k] + 1){
            b[k + 1]++;
            b[k] -= 3;
          }
        }
      }
    }
    return world;
  }

  // The next two functions are shortcut methods for disconnecting unwalkables

  public void disconnectUnwalkables(){
    for(int i = 0; i < nodes.size(); i++){
      Node temp = (Node)nodes.get(i);
      if(!temp.walkable){
        temp.disconnect();
      }
    }
  }

  public void radialDisconnectUnwalkables(){
    for(int i = 0; i < nodes.size(); i++){
      Node temp = (Node)nodes.get(i);
      if(!temp.walkable){
        temp.radialDisconnect();
      }
    }
  }

  // Utilities
  
  // Shortcut for dimSize on all nodes
  
  public void dimSize(int dim){
    for(int i = 0; i < nodes.size(); i++){
      Node n = (Node)nodes.get(i);
      n.dimSize(dim);
    }
  }

  // Returns an n-dimensional vector from a point on a line of units given an n-dimensional space

  public int [] getFolded(int n, int [] d){
    int [] coord = new int[d.length];
    for(int i = 0; i < d.length; i++){
      coord[i] = n;
      for(int j = d.length - 1; j > i; j--){
        int level = 1;
        for(int k = 0; k < j; k++){
          level *= d[k];
        }
        coord[i] %= level;
      }
      int level = 1;
      for(int j = 0; j < i; j++){
        level *= d[j];
      }
      coord[i] /= level;
    }
    return coord;
  }

  // Returns a point on a line of units from an n-dimensional vector in an n-dimensional space

  public int getUnfolded(int [] p, int [] d){
    int coord = 0;
    for(int i = 0; i < p.length; i++){
      int level = 1;
      for(int j = 0; j < i; j++){
        level *= d[j];
      }
      coord += p[i] * level;
    }
    return coord;
  }

}
