package aStarLibrary;

// AStar Node class, Aaron Steed 2006

import java.util.Vector;
import java.lang.Math;

public class Node{

  public float [] p;                        // Location, variable dimensions
  public Node parent = null;                // Parent Node setting
  public float f = 0;                       // Sum of goal and heuristic calculations
  public float g = 0;                       // Cost of reaching goal
  public float h = 0;                       // Heuristic distance calculation
  public Vector links = new Vector();       // Nodes attached to this Node
  public boolean walkable = true;           // Is this Node to be ignored?

  // Constructors

  public Node(float [] p) {
    this.p = p;
  }

  public Node(float [] p, Vector links){
    this.p = p;
    this.links = links;
  }

  // Field utilities

    public void reset() {
    parent = null;
    f = g = h = 0;
  }
  
  // Euclidean field methods

  public void setF(Node finish, boolean bfs) {
    setG();
    setH(finish);
    f = bfs ? h : g + h;
  }

  public void setG() {
    g = parent.g + dist(parent.p);
  }

  public void setH(Node finish) {
    h = dist(finish.p);
  }
  
  // Manhattan field methods - faster and designed for lack of corner nodes
  
  public void MsetF(Node finish, boolean bfs) {
    MsetG();
    MsetH(finish);
    f = bfs ? h : g + h;
  }

  public void MsetG() {
    g = parent.g + manhattan(parent.p);
  }

  public void MsetH(Node finish) {
    h = manhattan(finish.p);
  }

  // Linking tools

  public Node copy(){
    Vector temp = new Vector();
    temp.addAll(links);
    return new Node(copyP(), temp);
  }

  public void connect(Node n){
    links.add(n);
  }

  public void connect(Vector links){
    this.links.addAll(links);
  }

  // Iterates thru neighbours and unlinks them - Node is still linked to neighbours though

  public void disconnect(){
    for(int i = 0; i < links.size(); i++){
      Node n = (Node)links.get(i);
      int index = n.links.indexOf(this);
      if(index > -1){
        n.links.remove(n.links.indexOf(this));
      }
    }
  }

  // Calculates shortest link and kills all links around the Node in that radius
  // Used for making routes around objects account for the object's size

  public void radialDisconnect(){
    float radius = 0.0f;
    for(int j = 0; j < links.size(); j++){
      Node myLink = (Node)links.get(j);
      if(straightLink(myLink)){
        radius = dist(myLink.p);
        break;
      }
    }
    for(int j = 0; j < links.size(); j++){
      Node myLink = (Node)links.get(j);
      Vector removeMe = new Vector();
      for(int k = 0; k < myLink.links.size(); k++){
        Node myLinkLink = (Node)myLink.links.get(k);
        float [] midPoint = new float[myLinkLink.p.length > myLink.p.length ? myLinkLink.p.length : myLink.p.length];
        for(int m = 0; m < midPoint.length; m++){
          midPoint[m] = ((m < myLink.p.length ? myLink.p[m] : 0.0f) + (m < myLinkLink.p.length ? myLinkLink.p[m] : 0.0f)) * 0.5f;
        }
        if(dist(midPoint) <= radius){
          removeMe.add(myLinkLink);
        }
      }
      for(int k = 0; k < removeMe.size(); k++){
        Node temp = (Node)removeMe.get(k);
        int index = myLink.links.indexOf(temp);
        if(index > -1){
          myLink.links.remove(index);
        }
      }
    }
  }

  // Checks if a Node's position differs along one dimension only

  public boolean straightLink(Node myLink){
    if(links.indexOf(myLink) < 0){
      return false;
    }
    int directions = p.length > myLink.p.length ? p.length : myLink.p.length;
    int dimDelta = 0;
    for(int k = 0; k < directions; k++){
      if((k < p.length ? p[k] : 0) != (k < myLink.p.length ? myLink.p[k] : 0)){
        ++dimDelta;
      }
    }
    if(dimDelta == 1){
      return true;
    }
    return false;
  }

  // Location tools

  public float [] copyP(){
    float [] pTemp = new float[p.length];
    System.arraycopy(p, 0, pTemp, 0, p.length);
    return pTemp;
  }

  public float x(){
    if(p != null){
      return p[0];
    }
    return 0;
  }

  public float y(){
    if(p.length >= 2){
      return p[1];
    }
    return 0;
  }

  public float z(){
    if(p.length >= 3){
      return p[2];
    }
    return 0;
  }

  public float w(){
    if(p.length >=4){
      return p[3];
    }
    return 0;
  }
  
  public float [] addP(float [] b){
    float [] pTemp = new float[p.length > b.length ? p.length : b.length];
    for(int i = 0; i < pTemp.length; i++){
      pTemp[i] = (i < b.length ? b[i] : 0.0f) + (i < p.length ? p[i] : 0.0f);
    }
    return pTemp;
  }
  
  public float [] subP(float [] b){
    float [] pTemp = new float[p.length > b.length ? p.length : b.length];
    for(int i = 0; i < pTemp.length; i++){
      pTemp[i] = (i < b.length ? b[i] : 0.0f) - (i < p.length ? p[i] : 0.0f);
    }
    return pTemp;
  }

  public void dimSize(int dim){
    float [] pTemp = new float[dim];
    System.arraycopy(p, 0, pTemp, 0, p.length > pTemp.length ? pTemp.length : p.length);
    p = pTemp;
  }
  
  // Euclidean distance measuring for accuracy
  
  public float dist(float [] b) {
    switch(p.length){
    case 1:
      return
        ((p[0] - b[0]) < 0 ? (p[0] - b[0]) * -1 : (p[0] - b[0]));
    case 2:
      return
        (float)Math.sqrt(
      ((p[0] - b[0]) * (p[0] - b[0])) +
        ((p[1] - b[1]) * (p[1] - b[1])));
    case 3:
      return
        (float)Math.sqrt(
      ((p[0] - b[0]) * (p[0] - b[0])) +
        ((p[1] - b[1]) * (p[1] - b[1])) +
        ((p[2] - b[2]) * (p[2] - b[2])));
    }
    return
      (float)Math.sqrt(
    ((p[0] - b[0]) * (p[0] - b[0])) +
      ((p[1] - b[1]) * (p[1] - b[1])) +
      ((p[2] - b[2]) * (p[2] - b[2])) +
      ((p[3] - b[3]) * (p[3] - b[3])));
  }
  
  // Manhattan distance measuring for speed and avoiding jagged paths
  
  public float manhattan(float [] b) {
    switch(p.length){
    case 1:
      return
        ((p[0] - b[0]) < 0 ? -(p[0] - b[0]) : (p[0] - b[0]));
    case 2:
      return
        ((p[0] - b[0]) < 0 ? -(p[0] - b[0]) : (p[0] - b[0])) +
        ((p[1] - b[1]) < 0 ? -(p[1] - b[1]) : (p[1] - b[1]));
    case 3:
      return
        ((p[0] - b[0]) < 0 ? -(p[0] - b[0]) : (p[0] - b[0])) +
        ((p[1] - b[1]) < 0 ? -(p[1] - b[1]) : (p[1] - b[1])) +
        ((p[2] - b[2]) < 0 ? -(p[2] - b[2]) : (p[2] - b[2]));
    }
    return
      ((p[0] - b[0]) < 0 ? -(p[0] - b[0]) : (p[0] - b[0])) +
      ((p[1] - b[1]) < 0 ? -(p[1] - b[1]) : (p[1] - b[1])) +
      ((p[2] - b[2]) < 0 ? -(p[2] - b[2]) : (p[2] - b[2])) +
      ((p[3] - b[3]) < 0 ? -(p[3] - b[3]) : (p[3] - b[3]));
  }
}
